; Author: Dave Schlieder
;         10032 Pino Drive
;         Lakeside, Calif.  92040
;         (619)561-3770
;
;  Copyright 1994 Dave Schlieder
;  All rights reserved.
;
; Please read the accompanying QCNVRT.TXT file for additional information.
;
; This script contains a number of procedures that will allow a user to convert
; most Paradox saved queries to a PAL commanded query.
;
; These procedures have NO error handling in them.  The user may get a
; Cancel/Debug menu where the user might be restricted by access rights
; or when the file is already open, etc..
;
; You are free to use these procedures for your own use.  My only
; request is that you acknowledge me as the original author of these
; procedures and if you share this with anyone, that the Copyright and
; other information is left intact.
;
; You can simply play this script to demonstrate it's capabilities.  If you
; desire to add them to your libraries, you'll need to un-rem the appropriate
; WRITELIB statements and define the appropriate library name, etc. to add
; these to your libraries.


; AutoLib = "Paradox"


If Version() < 4.5 Then
   Beep
   Quit "Version 4.5 or higher of Paradox for DOS required"
EndIf

;
;                                                           QueryEndQueryToPAL()
;
; Purpose: Main procedure for the Query ... EndQuery to PAL command
;          conversion utility
Proc QueryEndQueryToPAL()
Private OutPutFile,
        InputFile,
        btn.Var,
        Choice,
        CheckPath,
        Editor.Bag,
        NumFound,
        QStart

PaintCanvas Fill "" Attribute 30 ALL
OutPutFile = Upper(PrivDir()+"INSTANT.SC")

While True

   ; PaintPAL_Generated_Code_Begin(485258446)

   ShowDialog "Query ... EndQuery Converter"

      @4,15 Height 12 Width 50

      ; PaintPAL_Frame_Begin
      Frame Single From 0,1 To 3,46
      PaintCanvas Attribute 127 0,1,0,46
      PaintCanvas Attribute 127 0,1,3,1
      PaintCanvas Attribute 112 3,2,3,46
      PaintCanvas Attribute 112 0,46,3,46
      ; PaintPAL_Frame_End

      ; PaintPAL_Frame_Begin
      Frame Single From 4,1 To 7,46
      PaintCanvas Attribute 112 4,1,4,46
      PaintCanvas Attribute 112 4,1,7,1
      PaintCanvas Attribute 127 7,2,7,46
      PaintCanvas Attribute 127 4,46,7,46
      ; PaintPAL_Frame_End

      Accept @6,3 Width 42
         "A65"
         Picture "*!"
         Tag "file.In"
         To InputFile

      Label @5,2
         "~I~nput File Name:"
         For "file.In"

      PushButton @8,9 Width 10
         "~O~K"
         Ok
         Default
         Value True
         Tag "btn.OK"
         To btn.Var

      PushButton @8,27 Width 10
         "~C~ancel"
         Cancel
         Value False
         Tag "btn.Cancel"
         To btn.Var

      Accept @2,3 Width 42
         "A65"
         Picture "*!"
         Tag "file.Out"
         To OutPutFile

      Label @1,2
         "~O~utput File Name:"
         For "file.Out"
   EndDialog

   ; PaintPAL_Generated_Code_End(485258446)

   If Not RetVal Then
      QUIT "Query ... EndQuery Converter Dialog Box canceled"
   EndIf

   Message "Checking input/output file parameters"

   Switch
      Case IsBlank( OutPutFile ) :
         Beep
         Message "Output file name is required!"
         Loop
      Case IsBlank( InputFile ) :
         Beep
         Message "Input file name is required!"
         Loop
      Case DirExists( InputFile ) = -1 :
         Beep
         Message "Input file name is invalid!"
         Loop
      Case DirExists( OutPutFile ) = -1 :
         Beep
         Message "Output file name is invalid!"
         Loop
   EndSwitch

   InputFile  = Upper(FullFileName(InputFile))
   OutPutFile = Upper(FullFileName(OutPutFile))

   Switch
      Case Not IsFile( InputFile ) :
         If Search( ".", InputFile ) = 0 Then
            InputFile = InputFile + ".SC"
         EndIf
         If Not IsFile( InputFile ) Then
            Beep
            Message "Input file name does not exist!"
            Loop
         EndIf
   EndSwitch

   Switch
      Case IsFile( OutPutFile ) :
         Beep
         Message "Output file already exists!"
         ShowMenu
            "Append"  : "Append to the existing file: " + OutPutFile ,
            "Replace" : "Replace the existing file: " + OutPutFile,
            "Esc"     : "Cancel the request and allow re-entry of output file name"
         To Choice
         Switch
            Case Choice = "Esc" :
               Loop
            Case Choice = "Replace" :
               Message "Deleting existing output file"
               Editor New OutPutFile
               CancelEdit
         EndSwitch
      Case Not IsFile( OutPutFile ) :
         ParseFileName OutPutFile To CheckPath
         If DirExists( CheckPath["Drive"] + CheckPath["Path"] ) <> 1 Then
            Beep
            Message "Invalid directory (", CheckPath["Drive"],
                    CheckPath["Path"], ") for Output file!"
            Loop
         EndIf
   EndSwitch

   ; Start an editor session with the InputFile
   Message "Opening and locating 1st occurrence of the keyword \"QUERY\" in the input file"
   Editor Open InputFile
   ; if this fails, then it is possible that the file is already open
   ; and needs to be addressed

   ; we can't have wordwrap on for this to work so we make sure it is off
   ; and need to reset it when we finish
   Editor Info To Editor.Bag
   If Editor.Bag["WORDWRAP"] Then
      {Options}
      {WordWrap}
      {Clear}
   EndIf

   ; initialize the find for the keyword "query"
   Editor Find "QUERY"

   If RetVal Then
      ; we found it but is it really the keyword query?
      Home  ; re-position the cursor to the beginning of the current line
            ; so the find next will find it again
      QStart = FindNextKeyWord("QUERY")
      ; if the keyword "QUERY" is not found on a line by itself,
      ; retval is set to false
      If QStart = 0 Then
         Retval = False
      Else
         Home ; re-position the cursor again
         RetVal = True
      EndIf
   EndIf

   ; if not found, then there is no saved query in the file
   If Not RetVal Then
      Beep
      If Editor.Bag["WORDWRAP"] Then
         {Options}
         {WordWrap}
         {Set}
      EndIf
      CancelEdit
      Message "Could not find the key word \"QUERY\" in the input file!"
      Loop
   EndIf

   ; return the cursor to the beginning of the current line
   Home

   NumFound = FindQuerySave()

   Message "Closing the input file"
   If Editor.Bag["WORDWRAP"] Then
      {Options}
      {WordWrap}
      {Set}
   EndIf
   CancelEdit
   If NumFound = 1 Then
      Message "Found ", NumFound, " saved query in the input file ", InputFile
   Else
      Message "Found ", NumFound, " saved queries in the input file ", InputFile
   EndIf

EndWhile

EndProc
;WriteLib AutoLib QueryEndQueryToPAL
;Release Procs    QueryEndQueryToPAL

;
;                                                                FindQuerySave()
;
; Purpose: Attempts to find all "QuerySave"s in the current editor window and
;          calls ConvertQuerySave to create the same query in PAL commands
Proc FindQuerySave()
Private ;QStart,   ; the start of the saved query
        QEnd,     ; the end of the saved query
        NumFound  ; number of saved queries found in the input file

NumFound = 0

; Until the matching keywords Query ... EndQuery are no longer in the file
While True
   ; attempt to find the matching end
   QEnd = FindNextKeyWord("ENDQUERY")
   ; if the keyword "ENDQUERY" is not found following the keyword "QUERY",
   ; the process is ended
   If QEnd = 0 Then
      Return NumFound
   EndIf
   NumFound = NumFound + 1
   ; now that we have found the matching "QUERY" ... "ENDQUERY" keywords,
   ; re-position the cursor on the line following the beginning of the
   ; saved query as we have determined above
   Editor GoTo Line QStart
   Down
   Home
   ; call the procedure to build the "PAL Command" query from this block of data
   ConvertQuerySave()
   ; re-initialize the find so that we can look for the next saved query in the
   ; file
   Editor GoTo Line QEnd
   Home
   Editor Find "QUERY"
   ; find the start of a saved query
   QStart = FindNextKeyWord("QUERY")
   ; if the keyword "QUERY" is not found on a line by itself,
   ; the process is ended
   If QStart = 0 Then
      Return NumFound
   EndIf
EndWhile

EndProc
;WriteLib AutoLib FindQuerySave
;Release Procs    FindQuerySave

;
;                                                             ConvertQuerySave()
;
; Purpose: Builds the query from the current saved query.  When this procedure
;          is called, the cursor is on the line after the keyword "query" which
;          marks the beginning of the saved query.
Proc ConvertQuerySave()
Private CurLine,        ; the current line of the editor window
        LastTable,      ; the name of the previous table found in the query
        CurTable,       ; the current table in the query
        OutFile,        ; the string to be appended to the output file
        NumLines,       ; the number of criteria rows for the current table
        QStruct,        ; the fields in the current table
        FirstField,     ; the field number for the next field found in the query
        CurFieldNum,    ; the current field number
        CurLineNum,     ; the current line number
        DSize,          ; a dynarray size variable
        AssignFieldNum, ; the current field number for assignment
        FieldAssign,    ; the value to be assigned to the current field
        FieldCheck,     ; the "CheckMarkStatus()" of the current field
        Editor.Bag      ; dynarray created with Editor Info Complete

; initialize the output variable
OutFile =
"; Input file: " + Upper(FullFileName(InputFile)) + "\n" +
"; \"Query\" found on line " + StrVal(QStart) + "\n" +
"; \"EndQuery\" found on line " + StrVal(QEnd) + "\n" +
"; \"PAL Command\" version of this QuerySave\n" +
"If NImages() > 0 Then\n" +
"   MoveTo 1\n" +
"   While ImageType() = \"Query\" ClearImage EndWhile\n" +
"EndIf\n\n"

; initialize the variable containing the name of the last table
LastTable = ""
; until we find the end of the query marked with "endquery"
While True
   ; move to the next non-blank/non-remark line in the query
   NextCodeLine()
   ; if the next non-blank line is the end of the query,
   ; we return from this procedure
   If Upper(CurLine) = "ENDQUERY" Then
      ; if we found at least one table, we append this information
      ; to the OutFile variable
      ; if we didn't find a table in this saved query, we append a
      ; message to the variable
      If LastTable <> "" Then
         AppendToOutFile()
      Else
         OutFile = OutFile +
         "; Failed to locate any table name in this Query ... EndQuery\n\n"
      EndIf
      ; write out what we know to the output file
      FileWrite Append OutPutFile From OutFile
      ; return and look for another saved query
      Return
   EndIf
   ; the first non-blank/non-remark line we find should contain the table
   ; name and fields for the first block of the query for that table.
   ; we can extract from that the table name and the names of the
   ; fields used in this block
   RetVal = Match(CurLine,"..|..",CurTable,CurLine)
   ; if this fails, then we have a query that this converter
   ; can not handle or is invalid as it con not determine a table
   ; name for that query on this current line
   If Not RetVal Then
      ; we ant to report the line number we failed on
      Editor Info Complete To Editor.Bag
      ; append this to the variable
      OutFile = OutFile  +
      "; Failed to locate a table in this Query ... EndQuery\n" +
      "; Line#: " + StrVal(Editor.Bag["CurLine"]) + "\n\n"
      ; write this information out
      FileWrite Append OutPutFile From OutFile
      ; and return and look for another query
      Return
   EndIf
   ; double up the backslashes as might be required
   CurTable = QuoteString(LTrim(RTrim(CurTable)))
   ; if this is a new table that we haven't found before
   If Upper(CurTable) <> Upper(LastTable) Then
      ; if this is not the first table we have found, we need to write
      ; out to the outfile variable the information we have on the
      ; current table before we re-initialize the variables
      If LastTable <> "" Then
         AppendToOutFile()
      EndIf
      Message "Working on query criteria for the table ", CurTable, " in the saved query"
      ; re-initialize the variables needed
      LastTable = CurTable ; so we know when we change tables
      DynArray QStruct[]   ; to keep track of the field names for the current table
      FirstField = 1       ; the first field we will find
      QStruct[0] = "#"     ; we keep track of values in this field (DELETE, etc.)
      NumLines = 0         ; keep track of the number of lines in the query
   Else
      ; re-set the first field counter for this block as this
      ; table has multiple blocks in the saved query
      FirstField = CurFieldNum + 1
   EndIf
   ; back this up one for the moment
   CurFieldNum = FirstField - 1
   ; we have found a table name, now we parse out the field names
   While Match(CurLine,"..|..",CurField,CurLine)
      ; for each field name we find, we up the counter by one
      ; and add the field name to the QStruct array
      ; we also apply quotestring to this though it probably
      ; isn't required as field names should not contain
      ; quotes and backslashes but...
      CurFieldNum = CurFieldNum + 1
      QStruct[CurFieldNum] = QuoteString(LTrim(RTrim(CurField)))
   EndWhile
   ; keep track of the current criteria line
   CurLineNum = 0
   While True
      Down
      Home
      ShiftPress "End"
      Editor Extract To CurLine
      ; trim off leading/trailing spaces
      CurLine = LTrim(RTrim(CurLine))
      ; if this line is meaningless, i.e. it's a remark line or blank or
      ; it's the end of the query, we quit from this while true loop
      If SubStr(CurLine,1,1) = ";" Or
         CurLine = "" Or
         Upper(CurLine) = "ENDQUERY" Then
         QuitLoop
      EndIf
      ; up the current criteria line counter by one
      CurLineNum = CurLineNum + 1
      ; adjust the total number of criteria lines to this value only
      ; if the CurLineNum is greater than that value
      NumLines = MAX(CurLineNum,NumLines)
      ; has the dynamic array QLine + StrVal(CurLineNum) been defined?
      DSize = DynArraySize( VarVal("QLine" + StrVal(CurLineNum) ) )
      ; if not, we create it and it's partner "QChecks"
      If DSize = 0 Then
         VarAssign DynArray "QLine" + StrVal( CurLineNum )
         VarAssign DynArray "QChecks" + StrVal( CurLineNum )
      EndIf
      ; the first match for a field assignment is always 0
      AssignFieldNum = 0
      While Match(CurLine,"..|..",FieldAssign,CurLine)
         ; add backslashes to quotes and backslashes as required
         FieldAssign = QuoteString(LTrim(RTrim(FieldAssign)))
         ; initialize this to a null
         FieldCheck = ""
         ; the following switch statement only applies to fields, not
         ; the column under the table name and only if the field starts
         ; with at least the word "check"
         ; if the entire field assignment is one of the check types or
         ; the field assignment starts with one of the check types followed
         ; by at least one space, we extract the check type into a
         ; separate variable and remove it from the fieldassign variable
         If AssignFieldNum <> 0 And Search("CHECK",FieldAssign) = 1 Then
            Switch
               Case Upper(FieldAssign) = "CHECK" :
                  FieldCheck = "Check"
                  FieldAssign = ""
               Case Search("CHECK ",FieldAssign) = 1 :
                  FieldCheck = "Check"
                  FieldAssign = LTrim(SubStr(FieldAssign,7,Len(FieldAssign)))
               Case Upper(FieldAssign) = "CHECKPLUS" :
                  FieldCheck = "CheckPlus"
                  FieldAssign = ""
               Case Search("CHECKPLUS ",FieldAssign) = 1 :
                  FieldCheck = "CheckPlus"
                  FieldAssign = LTrim(SubStr(FieldAssign,11,Len(FieldAssign)))
               Case Upper(FieldAssign) = "CHECKDESCENDING" :
                  FieldCheck = "CheckDescending"
                  FieldAssign = ""
               Case Search("CHECKDESCENDING ",FieldAssign) = 1 :
                  FieldCheck = "CheckDescending"
                  FieldAssign = LTrim(SubStr(FieldAssign,17,Len(FieldAssign)))
               Case Upper(FieldAssign) = "CHECKGROUPBY" :
                  FieldCheck = "GroupBy"
                  FieldAssign = ""
               Case Search("CHECKGROUPBY ",FieldAssign) = 1 :
                  FieldCheck = "GroupBy"
                  FieldAssign = LTrim(SubStr(FieldAssign,14,Len(FieldAssign)))
            EndSwitch
         EndIf
         ; assign the above values to the appropriate variables
         ; if we are on field 0, then there is the possibilty that we have already
         ; assigned a value to this element and we are about to null that value out
         If AssignFieldNum = 0 And
            IsAssigned(VarVal("QLine" + StrVal(CurLineNum),AssignFieldNum)) And
            VarVal("QLine" + StrVal(CurLineNum),AssignFieldNum) <> "" And
            FieldAssign = "" Then
            FieldAssign = VarVal("QLine" + StrVal(CurLineNum),AssignFieldNum)
         EndIf
         VarAssign "QLine" + StrVal(CurLineNum) Index AssignFieldNum From FieldAssign
         VarAssign "QChecks" + StrVal(CurLineNum) Index AssignFieldNum From FieldCheck
         ; if we just completed the row under the table name, we reset the
         ; variable to the first field number we will in the next match,
         ; otherwise we up the counter by one
         If AssignFieldNum = 0 Then
            AssignFieldNum = FirstField
         Else
            AssignFieldNum = AssignFieldNum + 1
         EndIf
      EndWhile
   EndWhile
EndWhile

EndProc
;WriteLib AutoLib ConvertQuerySave
;Release Procs    ConvertQuerySave

;
;                                                                 NextCodeLine()
;
; Purpose: Advances the cursor to the next non-blank/non-comment line of the
;          current file
Proc NextCodeLine()

While True
   ShiftPress "End"
   Editor Extract To CurLine
   If Search(";",CurLine) <> 0 Then
      CurLine = SubStr(CurLine,1,Search(";",CurLine)-1)
   EndIf
   CurLine = LTrim(RTrim(CurLine))
   ; if this is not a comment line and it is not blank, return
   If SubStr(CurLine,1,1) <> ";" And CurLine <> "" Then
      Return
   EndIf
   Down
   Home
EndWhile

EndProc
;WriteLib AutoLib NextCodeLine
;Release Procs    NextCodeLine

;
;                                                              AppendToOutFile()
;
; Purpose: Appends to the variable outfile the information we have gathered on
;          the current table in the saved query before advancing to the next
;          table in the query or after the end of the query has been found
;          The "IsAssigned" check on the dynamic arrays exists because not
;          every element of the array is necessarily filled in.  This can
;          happen on a saved query that has been edited, for example.
Proc AppendToOutFile()
Private CountOfLines,
        CountOfFields,
        Looper

Message "Writing out query criteria for the table ", CurTable, " in the saved query"
; create the {Ask} that will place the appropriate query image on the desktop
OutFile = OutFile + "{Ask} {" + LastTable + "}\n"
; for each criteria line for the current table
For CountOfLines From 1 To NumLines
   ; for each field that has some form of a "CheckMarkStatus()", create
   ; the code to moveto and place the appropriate check
   For CountOfFields From 1 To CurFieldNum
      If IsAssigned( VarVal( "QChecks" + StrVal(CountOfLines) , CountOfFields )) And
         VarVal( "QChecks" + StrVal(CountOfLines) , CountOfFields ) <> "" Then
         OutFile = OutFile +
         "MoveTo [" + QStruct[CountOfFields] + "] " +
         VarVal( "QChecks" + StrVal(CountOfLines) , CountOfFields ) + "\n"
      EndIf
   EndFor
   ; for each field that has some form of criteria, create the code
   ; to assign that criteria to the field
   For CountOfFields From 0 To CurFieldNum
      If IsAssigned( VarVal( "QLine" + StrVal(CountOfLines) , CountOfFields ) ) And
         VarVal( "QLine" + StrVal(CountOfLines) , CountOfFields ) <> "" Then
         OutFile = OutFile +
         "[" + QStruct[CountOfFields] + "] = \"" +
         VarVal( "QLine" + StrVal(CountOfLines) , CountOfFields ) + "\"\n"
      EndIf
   EndFor
   ; if there are more lines, we append the PAL commands to move the cursor
   ; down to the next row of the query image
   If NumLines > CountOfLines Then
      OutFile = OutFile + "Down\n"
   EndIf
EndFor
OutFile = OutFile + "If QueryOrder() = \"ImageOrder\" Then\n"
For CountOfFields From 1 To CurFieldNum
   OutFile = OutFile + "   MoveTo [" + QStruct[CountOfFields] + "] Rotate\n"
EndFor
If IsAssigned(QStruct[1]) Then
   OutFile = OutFile +
   "   CtrlHome Right\n" +
   "   While Upper(Field()) <> \"" + Upper(QStruct[1]) + "\" Rotate EndWhile\n"
EndIf
OutFile = OutFile +
"EndIf\n\n"
; re-define the dynamic arrays we created on the fly with this for loop
; to release the memory used by the current array assignments
; since you can't release vars varval and the other alternative
; is to use an execute statement, this is faster but not quite as
; efficient
For Looper From 1 To NumLines
   VarAssign DynArray "QLine"+StrVal(Looper)
   VarAssign DynArray "QChecks"+StrVal(Looper)
EndFor

EndProc
;WriteLib AutoLib AppendToOutFile
;Release Procs    AppendToOutFile

;
;                                                              FindNextKeyWord()
;
; Purpose: finds the start/end of a saved query. returns 0 if the keyword
;          specified in the parameter could not be found on a line by itself
;          (allows for a comment line following the keyword), otherwise
;          returns the line number that the word was found on.
;          assumes that the editor find value has already been specified
Proc FindNextKeyWord(KeyWord)
Private CurLine,
        Editor.Bag

Switch
   Case Upper(KeyWord) = "QUERY" :
      Message "Attempting to locate the beginning of a saved query"
   Case Upper(KeyWord) = "ENDQUERY" :
      Message "Attempting to locate the ending of a saved query"
   OtherWise :
      Message "Attempting to locate the keyword ", Upper(KeyWord)
EndSwitch

While True
   ; find the beginning/end of the saved query from the current cursor position
   ; in the file
   Editor FindNext
   ; if not found, then return 0
   If Not RetVal Then
      Return 0
   EndIf
   ; return the current line number if the keyword and only the keyword
   ; (less comments) is on the current line, otherwise we'll loop and try the
   ; findnext again
   Home
   ShiftPress "End"
   Editor Extract To CurLine
   If Search(";",CurLine) <> 0 Then
      CurLine = SubStr(CurLine,1,Search(";",CurLine)-1)
   EndIf
   CurLine = LTrim(RTrim(CurLine))
   If Upper(CurLine) = Upper(KeyWord) Then
      Editor Info Complete To Editor.Bag
      Return Editor.Bag["CurLine"]
   EndIf
EndWhile

EndProc
;WriteLib AutoLib FindNextKeyWord
;Release Procs    FindNextKeyWord

QueryEndQueryToPAL()
